#include "include.h"

static void Terminate (int)
{
	if (getpid () == Globals.MasterPID) {
		EINTR_Valid = true;
		Globals.Shutdown = true;
	}
}

once Globals;

// settings registry file
#define STTI_AUTH	"author="
#define STTI_SELF	"myaddr="
#define STTI_LOCP	"lp="
#define STTI_FIRE	"firewalled="
#define STTI_WFIN	"reply_to_finger="
#define STTI_QSLP	"queuersleep="
#define STTI_HTTP	"httpport="
#define STTI_HTPP	"httppass="

#define ISOPTION(x,y)	(!strncmp (x, y, sizeof y - 1))

const char noopt [] = "Can't open settings file, run 'postnet -c' to create\n";
const char nocri [] = "Settings file does not have 'myaddress' or 'myname', run 'postnet -c' to create";

bool once::Initialization ()
{
	FILE *f;
	char line [300];

	ThisAuthor = ThisPoint = NULL;
	QueuerSleep = 60;
	Shutdown = false;
	nClients = 0;
	HTTPPort = 0;
	HTTPPassword = NULL;

	// first load from the settings file
	if ((f = fopen (SETTINGS_FILE, "r"))) {
		while (fgets (line, 299, f))
		if (ISOPTION(line, STTI_AUTH)) {
			line [strlen (line) - 1] = 0;
			ThisAuthor = StrDup (line + sizeof STTI_AUTH - 1);
		} else if (ISOPTION(line, STTI_SELF)) {
			line [strlen (line) - 1] = 0;
			ThisPoint = StrDup (line + sizeof STTI_SELF - 1);
		} else if (ISOPTION(line, STTI_LOCP)) {
			line [strlen (line) - 1] = 0;
			LOCpass = StrDup (line + sizeof STTI_LOCP - 1);
		} else if (ISOPTION(line, STTI_HTPP)) {
			line [strlen (line) - 1] = 0;
			HTTPPassword = StrDup (line + sizeof STTI_HTPP - 1);
		} else if (ISOPTION(line, STTI_QSLP))
			QueuerSleep = strtoul (line + sizeof STTI_QSLP - 1, NULL, 10);
		else if (ISOPTION(line, STTI_HTTP))
			HTTPPort = strtoul (line + sizeof STTI_HTTP - 1, NULL, 10);
		else if (ISOPTION(line, STTI_FIRE))
			AmIFirewalled = line [sizeof STTI_FIRE - 1] == 'y';
		else if (ISOPTION(line, STTI_WFIN))
			CanReplyToFinger = line [sizeof STTI_WFIN - 1] == 'y';
		fclose (f);
	} else {
		fputs (noopt, stderr);
		return false;
	}

	if (!ThisAuthor || !ThisPoint) {
		fputs (nocri, stderr);
		return false;
	}
	if (QueuerSleep < 60) QueuerSleep = 60;
	if (!HTTPPassword) {
		HTTPPassword = StrDup ("foo");
		printf ("HTTP password not set. "
			"Use `foo' to login and change it from /config\n");
	}

	// POSTNET protocol words. All clients must argee on the strings.
	PMSGtext [PVERSION]		= "PNET";
	PMSGtext [DOWNLOAD]		= "DLD";
	PMSGtext [NOT_AVAILABLE]	= "N/A"; 
	PMSGtext [PNET_OK]		= "OK ";
	PMSGtext [GET]			= "GET";
	PMSGtext [HAVE_REPLY]		= "IHR";
	PMSGtext [YOU_HAVE_REPLY]	= "YHR";
	PMSGtext [POINT_CHANGE_REPLY]	= "PCR";
	PMSGtext [VERIFY_R]		= "VCR";
	PMSGtext [POINT_CHANGE_PARENT]	= "PCP";
	PMSGtext [VERIFY_P]		= "VCP";
	PMSGtext [REDONE]		= "RED";
	PMSGtext [PING]			= "PING";

	// if they hadn't screwed enum with operator
	// overloading, i'd be using enums.
	PNetFields [PFIELD_TITLE]	= "title=";
	PNetFields [PFIELD_AUTHOR]	= "author=";
	PNetFields [PFIELD_DATE]	= "date=";
	PNetFields [PFIELD_CRC]		= "crc=";

	fieldOverhead = sizeof ("title=") + sizeof ("author=") +
			sizeof ("date=") + sizeof ("crc=");

	// create the server greeting message
	char tmp [400];
	o_argstream O (tmp, 400);
	O.putarg (PMSGtext [PVERSION]);
	O.putarg (POSTNET_VERSION);
	O.putarg (ThisAuthor);
	O.putarg (ThisPoint);

	Greeting.Text = new char [Greeting.size = O.have];
	memcpy (Greeting.Text, O.data, O.have);

	// initializations
	//LOG = fopen (LOGFILE, "a");
	sem_init (&DatabaseLock, 0, 0);
	sem_init (&QueuerLock, 0, 1);
	RDiff.c = 0;
	RDiff.last = 0;
	RecoveryFd = 0;
	nPS = nPR = 0;

	umask (0177);
	srand (time (NULL));

	MasterPID = getpid ();
	signal (SIGTERM, Terminate);
	signal (SIGINT, Terminate);

	return true;
}

void once::SaveSettings ()
{
	FILE *f = fopen (SETTINGS_FILE, "w");
	if (!f) {
		fprintf (stderr, "Can't save settings!\n");
		return;
	}
	fputs (BB BLI "Do NOT edit this file. Use postnet -c\n", f);
	fprintf (f, STTI_AUTH"%s\n", ThisAuthor);
	fprintf (f, STTI_SELF"%s\n", ThisPoint);
	if (LOCpass)
	fprintf (f, STTI_LOCP"%s\n", LOCpass);
	fprintf (f, STTI_HTPP"%s\n", (HTTPPassword) ? HTTPPassword : "foo");
	fprintf (f, STTI_QSLP"%i\n", QueuerSleep);
	fprintf (f, STTI_HTTP"%i\n", HTTPPort);
	fprintf (f, STTI_FIRE"%c\n", (AmIFirewalled)?'y':'n');
	fprintf (f, STTI_WFIN"%c\n", (CanReplyToFinger)?'y':'n');
	fclose (f);
}

void logprintf (int level, const char *fmt, ...)
{
	if (level < Globals.loglevel && Globals.LOG) {
		va_list args;
		va_start (args, fmt);
		// this is not atomic, we risk?
		fprintf (Globals.LOG, "[%i] ", getpid ());
		vfprintf (Globals.LOG, fmt, args);
		fflush (Globals.LOG);
		va_end (args);
	}
}

PNETMSG PMSGtype (char *s)
{
	int i;

	for (i = (int)PVERSION; i < BAD_MESSAGE; i++)
		if (!strcmp (s, Globals.PMSGtext [i])) return (PNETMSG)i;

	return (PNETMSG)i;
}

static void forkdaemon ()
{
	switch (fork ()) {
	case 0: Globals.MasterPID = getpid (); return;
	case -1: fputs ("Can't fork in the background!\n", stderr);
	default: exit (0);
	}
}

void ShutDown ()
{
	// shutdown means polls must be interrupted.
	// since this is multithreaded i don't know
	// if Terminate() will be called before polls
	// return with -1&&EINTR
	// for internal termination (this function) we
	// are OK. For external termination (Ctrl-C or
	// shell kill) we hope for the best.
	EINTR_Valid = true;
	Globals.Shutdown = true;
	kill (Globals.MasterPID, SIGTERM);
}




static const char help [] = "\
postnet 1.0\n\
[implements POSTNET protocol specification ["POSTNET_VERSION"],\
 compiled ["__DATE__"]]\n\
\n\
Written by Stelios Xanthakis <axanth@tee.gr>\n\
\n\
Usage: postnet [-c] | [-E <file.ipi>] | \n\
	       [[-d <data-directory>] [-i <user-interface>] [-l<log-level>]\n\
\n\
	-i hl		: http interface on localhost only (default)\n\
	-i h<IP>	: http interface to remote host IP\n\
	-i r		: raw user interface (for debugging)\n\
	-i hc		: try another TCP port for the http server\n\
			  if the standard one is not available\n\
\n\
	-c		: edit settings\n\
	-E <file.ipi>	: manually edit post files\n\
";







static char enterauthor [] = "(nick)name not set. Please enter your POSTNET name";
static char cpprompt [] = "\
Enter your contact address. This is:\n\
- if you have a static IP or dyndns (you are always reachable at a standard\n\
  internet hostname), the address will be:\n\
     /dns/hostname/port\n\
  where <port> a number of your choice between 2000 and 30000.\n\
  For example:   /dns/voyager.di.auo.com/11000\n\
- if you have a LOC account the address will be:\n\
     /loc/locserver_address/locserver_port/loc_username\n\
  For example:   /loc/freeloc.org/9090/jbop231\n\
\n\
What will it be";
static char frwlq [] = "Are you firewalled (y/n)?";
static char passprompt [] = "Enter LOC account password";





static bool goodaddress (char *a)
{
	if (strlen (a) > MAX_PRL - 16) {
		printf ("You will have trouble with such a long address\n");
		return false;
	}
	if (!strncasecmp (a, "/dns/", 5)) {
		if (slashcount (a) != 3) return false;
		if (strtol (strrchr (a, '/')-1, NULL, 10) == 0) return false;
		return true;
	} else if (!strncasecmp (a, "/loc/", 5)) {
		if (slashcount (a) != 4) return false;
		char c [12];
		do {
			if (!getuserline (frwlq, c, 5)) exit (1);
		} while (*c != 'y' && *c != 'n' && *c != 'Y' && *c != 'N');
		Globals.AmIFirewalled = *c == 'y' || *c == 'Y';
		while (!getuserline (passprompt, c, 12));
		Globals.LOCpass = StrDup (c);
		return true;
	}
	printf ("What's that?\n");
	return false;
}

static void configure ()
{
	char oa [MAX_AUTHOR], op [MAX_PRL], *p;

	Globals.Initialization ();
	if (Globals.ThisAuthor == NULL)
		if (getuserline (enterauthor, oa, MAX_AUTHOR))
			Globals.ThisAuthor = StrDup (oa);
		else exit (1);

	do {
		if (Globals.ThisPoint)
			printf ("Current Address: %s\n", Globals.ThisPoint);
		if (!getuserline (cpprompt, op, MAX_PRL))
			exit (1);
	} while (!goodaddress (op));

	p = Globals.ThisPoint;
	Globals.ThisPoint = StrDup (op);

	if (p && strcasecmp (Globals.ThisPoint, p)) {
		LoadIPI ();
		PrepareRecovery ();
		myPointChange ();
		SaveIPI ();
	}

	Globals.SaveSettings ();
}

static void parse (int argc, char **argv)
{
	char *home = NULL, dhome [] = POSTNET_HOME;
	char *ed = NULL, buffer [16];
	bool create = false;
	int i;

	Globals.iOptions.DoRaw = Globals.iOptions.DoHttp = false;
	Globals.iOptions.FreePort = false;
	Globals.iOptions.VIP = LOCALHOST;
	for (i = 0; i < argc; i++)
		if (!strcmp (argv [i], "-d") && i+1 < argc)
			home = argv [++i];
		else if (!strcmp (argv [i], "-c"))
			create = true;
		else if (!strcmp (argv [i], "-l"))
			Globals.loglevel = strtol (&argv [i][1], NULL, 10);
		else if (!strcmp (argv [i], "-i")) {
			if (argv [++i][0] == 'r') Globals.iOptions.DoRaw = true;
			else if (argv [i][0] == 'c')
				Globals.iOptions.FreePort = true;
			else if (argv [i][0] == 'h') {
				Globals.iOptions.DoHttp = true;
				if (argv [i][1] != 'l')
					Globals.iOptions.VIP =
						nslookup (argv [i] + 1);
			} else goto badopt;
		} else if (!strcmp (argv [i], "-E"))
			ed = argv [++i];
		else if (argv [i] [0] == '-') {
badopt:
			puts (help);
			exit (0);
		}

	if (!Globals.iOptions.DoRaw && !Globals.iOptions.DoHttp)
		Globals.iOptions.DoHttp = true;

	if (!home) {
		if (dhome [0] == '~') {
			if (!getenv ("HOME")) {
				fputs ("Can't find $HOME\n", stderr);
				exit (1);
			}
			home = (char*) alloca (200);
			strcat (strcpy (home, getenv ("HOME")), dhome + 1);
		} else home = dhome;
	}

	if (chdir (home) == -1) {
		fprintf (stderr, "Can't go home \"%s\"\n", home);
		exit (1);
	}

	if (!access (LOCK_FILE, F_OK)) {
		printf ("Lockfile ["LOCK_FILE"] found.\n"
			"Make sure postnet is not already running\n");
		exit (1);
	}

	if (ed) {
		EditIPI (ed);
		exit (0);
	}

	if (create) {
		configure ();
		exit (0);
	}

	sprintf (buffer, "%i\n", getpid ());
	i = open (LOCK_FILE, O_WRONLY | O_CREAT);
	write (i, buffer, strlen (buffer));
	close (i);
}

int main (int argc, char **argv)
{
	parse (argc - 1, argv + 1);

	if (!Globals.iOptions.DoRaw) forkdaemon ();

	// import settings
	if (!Globals.Initialization ()) return 1;

	// load posts
	LoadIPI ();

	// if recovery do it and save recovered posts
	PrepareRecovery ();

	// scan database for things that need doing
	QinitDB ();

	// load banned
	LoadBanned ();

	// start queuer thread
	QStart ();

	// start user interfaces
	if (Globals.iOptions.DoRaw) forkraw ();

	if (Globals.iOptions.DoHttp)
		forkhttp (Globals.iOptions.VIP, Globals.iOptions.FreePort);

	// start server
	if (!Globals.Shutdown) enter_POSTNET ();

	fprintf (stderr, "POSTNET terminated\n");

	// Terminate
	CloseRecovery ();
	SaveBanned ();
	Globals.SaveSettings ();
	if (Globals.LOG) fclose (Globals.LOG);
	unlink (LOCK_FILE);
}

#ifdef EFENCE
// ###### Electric Fence ######
// These activate efence on our C++ allocations.
// #############################################

void *operator new (size_t s)
{
	return malloc (s);
}

void operator delete (void *p)
{
	free (p);
}

void *operator new[] (size_t s)
{
	return malloc (s);
}

void operator delete[] (void *p)
{
	free (p);
}

#endif
